/**
 *  Copyright (C) 2005 PGT-Linux.org http://www.pgt-linux.org
 *  Author: vandorpe Olivier <vandorpeo@pgt-linux.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <usb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <pthread.h>
#include <syslog.h>
#include "usbb2k-api.h"

#define USBB2K_SOCKET "/tmp/usbb2k.sock"
#define USBB2K_PID    "/var/run/usbb2k_api.pid"

void
signal_handler(int signal)
{
  switch(signal)
  {
    case 1: /* SIGHUP */
      syslog(LOG_INFO,"SIGHUP receive");
      break;
    case 15:/* SIGTERM */
      syslog(LOG_INFO,"SIGTERM receice");
      if (usb_loop==OFF) 
      {
        closelog();
	exit(1);
      }
      usb_loop=OFF;
      main_loop=OFF;
      closelog();
      break;
  }
}

usb_dev_handle *
usbb2k_init(struct usb_device* dev)
{
  int err;
  char data[URB_LENGTH];
  usb_dev_handle *dev_h;

  if (!(dev_h=usb_open(dev))) 
  {
    return NULL;	
  }
  err=usb_claim_interface(dev_h,CONTROL);
  if(err != 0)
  {
    usb_detach_kernel_driver_np(dev_h, CONTROL);
    err=usb_claim_interface(dev_h,CONTROL);
    if (err !=0)
    {
      syslog(LOG_ERR,"claim interface Failed");
      usb_close(dev_h);
      return NULL;
    }
  }

  err=usb_control_msg(dev_h,USB_MSG_OUT,0x9,0x200,0x03,URB_CMD[URB_INIT],URB_LENGTH,TIME_OUT);
  if (err!=URB_LENGTH) return NULL;
  err=usb_interrupt_read(dev_h,0x81,data,URB_LENGTH,TIME_OUT);
  if (err!=URB_LENGTH) return NULL;
  return dev_h;
}

int 
usbb2k_check_handset(usb_dev_handle *dev_h)
{
  int err;
  char data[URB_LENGTH]; 
  err=usb_control_msg(dev_h,USB_MSG_OUT,0x9,0x200,0x3,URB_CMD[URB_HANDSET],URB_LENGTH,TIME_OUT);
  if (err!=URB_LENGTH) 
  {
    syslog(LOG_ERR,"error send check_handset:%d [%x]",err,dev_h);
  }
  err=usb_interrupt_read(dev_h,0x81,data,URB_LENGTH,TIME_OUT);
  if (err==URB_LENGTH) 
  {
    if (data[4]==0x02) 
    {
      return ON;
    }
    else
    {
      return OFF;
    }
  }
  syslog(LOG_ERR,"error: %d",err);
  return ERR;
}

int 
usbb2k_check_keypress(usb_dev_handle *dev_h)
{
  int err;
  char data[URB_LENGTH];  
  err=usb_control_msg(dev_h,USB_MSG_OUT,0x9,0x200,0x3,URB_CMD[URB_KEYPRESS],URB_LENGTH,TIME_OUT);
  err=usb_interrupt_read(dev_h,0x81,data,URB_LENGTH,TIME_OUT);
  if (err==URB_LENGTH) {
    return data[4];
  }
  return -1;
}

int 
data_copy(unsigned char * buf,unsigned char *orig,int l)
{
  int i;
  for(i=0;i<l;i++) buf[i]=orig[i];
  return i;
}

int 
usbb2k_get_key(usb_dev_handle *dev_h,int keynum)
{
  int err;
  char data[URB_LENGTH];  
  keynum--;
  if (keynum<0) keynum=0x7E;
  data_copy(data,URB_CMD[URB_KEYCODE],URB_LENGTH);
  data[3]=keynum;
  data[15]=0x7e - keynum;
  err=usb_control_msg(dev_h,USB_MSG_OUT,0x9,0x200,0x3,data,URB_LENGTH,TIME_OUT);
  err=usb_interrupt_read(dev_h,0x81,data,URB_LENGTH,TIME_OUT);
  if (err==URB_LENGTH) 
  {
    //if (data[0]!=URB_CMD[URB_KEYCODE][0]) return -2;
    return data[4];
  }
  return err;
}

struct usb_device* 
scan_bus()
{
  struct usb_bus *bus;
  struct usb_device *device;
  usb_init();
  usb_find_busses();
  usb_find_devices();
  
  for(bus=usb_busses;bus;bus=bus->next)
  {
    for (device = bus->devices;device;device=device->next)
    {
      if ((device->descriptor.idVendor == telbox_idVendor) && 
          (device->descriptor.idProduct ==telbox_idProduct)) 
      {
        return device;
      }
    }
  }	
  return NULL;
}

int 
buffcmp(const char *buf1,const char *buf2)
{
  int l,i;	
  if (strlen(buf2)>strlen(buf1)) return -1;
  l=strlen(buf2);
  i=0;
  while(i<l)
  {
    if (buf1[i]!=buf2[i]) return i+1;
    i++;
  }
  return 0;
}

int 
decode_cmd(const char *cmd_str,char *arg)
{
  if (buffcmp(cmd_str,"SWITCH U")==0) 
  {
    return CMD_SWITCH_USB;
  }

  if (buffcmp(cmd_str,"SWITCH P")==0) 
  {
    return CMD_SWITCH_PSTN;
  }
  if (buffcmp(cmd_str,"RING")==0)
  {
    if (strlen(cmd_str)<5)
    {
      sprintf(arg,"0");
      return CMD_RING;
    }
    sprintf(arg,"%s",cmd_str+5);	
    return CMD_RING;
  }
  if (buffcmp(cmd_str,"Q")==0)
  {
    return CMD_QUIT;
  }
  // --- added by ant@loadtrax.com 30/10/2005 ---
  if (buffcmp(cmd_str,"TONE")==0)
  {
    sprintf(arg,"%s",cmd_str+5);
    return CMD_TONE;
  }
  // ---
  return CMD_NA;
}

int 
usbb2k_tone(usb_dev_handle *dev_h,int status)
{
  int err;
  char data[URB_LENGTH];
  if (status==ON) 
  {
    err=usb_control_msg(dev_h,USB_MSG_OUT,0x9,0x200,0x03,URB_CMD[URB_TONE_ON],URB_LENGTH,TIME_OUT);
    err=usb_interrupt_read(dev_h,0x81,data,URB_LENGTH,TIME_OUT);
    return ON;
  }
  else
  {
    err=usb_control_msg(dev_h,USB_MSG_OUT,0x9,0x200,0x03,URB_CMD[URB_TONE_OFF],URB_LENGTH,TIME_OUT);
    err=usb_interrupt_read(dev_h,0x81,data,URB_LENGTH,TIME_OUT);
    return OFF;
  }
}

int 
usbb2k_ring(usb_dev_handle *dev_h,int status)
{
  int err;
  char data[URB_LENGTH];

  if (status==ON) 
  {
    err=usb_control_msg(dev_h,USB_MSG_OUT,0x9,0x200,0x03,URB_CMD[URB_RING_ON],URB_LENGTH,TIME_OUT);
    err=usb_interrupt_read(dev_h,0x81,data,URB_LENGTH,TIME_OUT);
    return ON;
  }
  else
  {
    err=usb_control_msg(dev_h,USB_MSG_OUT,0x9,0x200,0x03,URB_CMD[URB_RING_OFF],URB_LENGTH,TIME_OUT);
    err=usb_interrupt_read(dev_h,0x81,data,URB_LENGTH,TIME_OUT);
    return OFF;
  }
}


int 
usbb2k_switch_mode(usb_dev_handle *dev_h,int mode)
{
  int err;
  char data[URB_LENGTH];
  if (mode==USB_MODE)
  {
    err=usb_control_msg(dev_h,USB_MSG_OUT,0x9,0x200,0x03,URB_CMD[URB_LED_USB_ON],URB_LENGTH,TIME_OUT);
    err=usb_interrupt_read(dev_h,0x81,data,URB_LENGTH,TIME_OUT);
    err=usb_control_msg(dev_h,USB_MSG_OUT,0x9,0x200,0x03,URB_CMD[URB_USB_ON],URB_LENGTH,TIME_OUT);
    err=usb_interrupt_read(dev_h,0x81,data,URB_LENGTH,TIME_OUT);
    return USB_MODE;
  }
  if (mode==PSTN_MODE)
  {
    err=usb_control_msg(dev_h,USB_MSG_OUT,0x9,0x200,0x03,URB_CMD[URB_LED_USB_OFF],URB_LENGTH,TIME_OUT);
    err=usb_interrupt_read(dev_h,0x81,data,URB_LENGTH,TIME_OUT);
    err=usb_control_msg(dev_h,USB_MSG_OUT,0x9,0x200,0x03,URB_CMD[URB_USB_OFF],URB_LENGTH,TIME_OUT);
    err=usb_interrupt_read(dev_h,0x81,data,URB_LENGTH,TIME_OUT);
    return PSTN_MODE;
  }
  return -1;
}
//
// Boucle sur la sonnerie tant que (ou):
// - annulation de la sonnerie (cmd RING 0)
// - detection du decrochage du combiné.
// - perte de la connexion client
// - timeout = 0
void *
ring_th(void *arg) {
  int err,timeout,ringcycle,ringcyclelength;
  char thissegment;

  timeout=30;
  syslog(LOG_INFO,"Ring Thread...on");
  ringing_state=ON;
  while(timeout>0 && ringing_state==ON)
  {
    timeout--;
    // --- added by ant@loadtrax.com 30/10/2005 ---
    // ringing_tone[] contains, eg, DbDt,  where letter
    // A/a=0.1s, Y/y=2.5s, UPPER=ring, lower=no ring
    // Z/z is 'stop here', ie, only do one ring loop
    // Try AaAaAx :-)
    ringcyclelength=strlen(ringing_tone);
    if (ringcyclelength<1) {
	// the user didn't set a valid tone, so apply a default. DbDt is
	// the UK ringing tone
        sprintf(ringing_tone,"DbDt");
	ringcyclelength=strlen(ringing_tone);
    }
    for (ringcycle=0;ringcycle<ringcyclelength;ringcycle++) {
      thissegment=ringing_tone[ringcycle];
      ringing_mode=(thissegment & 0x20)?OFF:ON;
      if ((thissegment & 0x1f)==26) {
	// user specified 'z' or 'Z' so drop out here
	timeout=0;
	ringcycle=ringcyclelength;
        ringing_mode=OFF;
	ringing_state=OFF;
      }
      else usleep((thissegment & 0x1f)*100000);
    }
    // ---
  }
  syslog(LOG_INFO,"Ring Thread..off");
  pthread_exit(NULL);
} 

// Wait client instruction
void *
client_th(void *client_arg){
  int msgl,ring_type;
  int client,ret;
  usb_dev_handle *dev_h;
  char msg[MSG_MAXL+1],arg_cmd[MSG_MAXL+1];
  pthread_t ring_thread; 
  
  dev_h=((client_arg_t*) client_arg)->dev_h;
  client=((client_arg_t*) client_arg)->sock;
  do 
  {
    msgl=read(client,msg,MSG_MAXL);
    if (msgl>1) {
      //traite CMD
      msg[msgl-1]=0;
      syslog(LOG_INFO,"RECV CMD:[%s]",msg);
      switch(decode_cmd(msg,arg_cmd))
      {
        case CMD_SWITCH_USB:
	  usbb2k_switch_mode(dev_h,USB_MODE);
	  break;
	case CMD_SWITCH_PSTN:
	  usbb2k_switch_mode(dev_h,PSTN_MODE);
	  break;
	case CMD_QUIT:
	  msgl=-2;
	  break;
	case CMD_RING:
	  ring_type=atoi(arg_cmd);
	  if (ring_type>0 && ringing_state==OFF && handset_state==OFF) {
	    syslog(LOG_INFO,"Ring init_thread");
	    if ((ret=pthread_create(&ring_thread,NULL,ring_th,NULL))!=0){
	      pthread_detach(ring_thread);
	      syslog(LOG_ERR,"ring thread error");
	      closelog();
	    }
	  }
	  else
	  {
	    ringing_state=OFF;
	    ringing_mode=OFF;
	  }
	  break;
	// --- added by ant@loadtrax.com 30/10/2005 ---
	case CMD_TONE:
	  // copy arg_cmd[] into ringing_tone[]
	  sprintf(ringing_tone,"%s",arg_cmd);
	  
	  break;
	// ---
	case CMD_NA:
	default:
	  syslog(LOG_INFO,"commande not def");
      }
    }
  }while(msgl>0);
}


/* USB thread */
void *
usblect_th(void *client_arg)
{
  int client;
  int keypress_state,tone_flag;
  int handset_state_flag,keypress_state_flag,key_code;
  int ringing_mode_flag;
  usb_dev_handle *dev_h;
  int msgl;
  char msg[MSG_MAXL+1],arg_cmd[MSG_MAXL+1];

  dev_h=((client_arg_t*) client_arg)->dev_h;
  client=((client_arg_t*) client_arg)->sock;
  syslog(LOG_INFO,"Thread USB");
  /* init */
  ringing_mode_flag=ERR;
  handset_state_flag=ERR;
  keypress_state_flag=ERR;
  usb_loop=ON; 
  /* loop */ 
  keypress_state=usbb2k_check_keypress(dev_h);
  keypress_state_flag=keypress_state;

  while(usb_loop==ON) 
  {
    /* Handset check */
    handset_state=usbb2k_check_handset(dev_h);
    if (handset_state==ERR) syslog(LOG_ERR,"handset Error");
    if (handset_state_flag!=handset_state) 
    {
      handset_state_flag=handset_state;
      if (handset_state==ON) 
      { /* DECROCHE */
        sprintf(msg,"HANDSET ON \n");
        if (ringing_state==ON) 
	{
	  ringing_state=OFF;
	  ringing_mode=OFF;
	  ringing_mode_flag=OFF;
	  usbb2k_ring(dev_h,OFF);
	}
	else
	{
          tone_flag=usbb2k_tone(dev_h,ON);
	}
      }
      else 
      { /* RACROCHE */
        sprintf(msg,"HANDSET OFF\n");
	if (tone_flag==ON)
	{
          tone_flag=usbb2k_tone(dev_h,OFF);
	}
      }
      msgl=write(client,msg,strlen(msg));
      if (msgl<strlen(msg)) 
      {
        msgl=-2;
        continue;
       }
     }

    if (handset_state==ON)
    { /* KEY PRESS check */
       keypress_state=usbb2k_check_keypress(dev_h);
       if (keypress_state_flag!=keypress_state) 
       {
         keypress_state_flag=keypress_state;
         if ((key_code=usbb2k_get_key(dev_h,keypress_state))!=ERR) 
	 {
           sprintf(msg,"KEY  %.2x\n",key_code);
           msgl=write(client,msg,strlen(msg));		
	 }
	 else syslog(LOG_ERR,"Key ERROR %d : %d ",key_code,keypress_state);
         if (tone_flag==ON) 
	 {
             tone_flag=usbb2k_tone(dev_h,OFF);
         }
       }
     }
     else
     {
       if (ringing_state==ON || ringing_mode_flag==ON)
       { /* RINGING MODE */
         if (ringing_mode_flag!=ringing_mode)
	 {
	   ringing_mode_flag=ringing_mode;
	   if (ringing_mode_flag==ON)
	   {
	     usbb2k_ring(dev_h,ON);
	   }
	   else
           {
	     usbb2k_ring(dev_h,OFF);
	   }
	 }
       } 
     }
     //usleep(100);
  }
  /* usb loop stop -> close client */
  sprintf(msg,"USB DISCONNECT\n");
  msgl=write(client,msg,strlen(msg));
  /* Stop Ringing LOOP */
  ringing_state=OFF;
  /* Stop ring */
  usbb2k_ring(dev_h,OFF);
  
  if (main_loop==OFF) 
  {
    sprintf(msg,"USBB2K OFF\n");
    msgl=write(client,msg,strlen(msg));
    syslog(LOG_INFO,"shutdown");
    close(client);
  }
  pthread_exit(NULL);
}

int
daemon_mode()
{ /*  DAEMON MODE */
  int pid;
  int i,fic;
  char pid_file[PID_MAXL];
  char pid_str[10];


  /* Fork process */
  pid=fork();
  if (pid<0)
  {
    perror("echec fork");
    exit(-1);
  }
  if (pid>0) 
  {
    exit(0);
  }

  if (setsid()<0)
  {
    perror("echec setsid");
    exit(-1);
  }
  
  /* Syslog */
  openlog("usbb2k_api",LOG_PID,LOG_DAEMON);
  syslog(LOG_INFO,"usbb2k init ok");
  
  for(i=getdtablesize()-1;i>=0;i--)
  {
    close(i);
  }
  /* flux stdin */
  i=open("/dev/null",O_RDWR);
  /* flux stdout */
  dup(i);
  /* flux stderr */
  dup(i);
  
  /* the process don't create executable file */
  umask(0222);

  sprintf(pid_file,"%s",USBB2K_PID);
  /* Check if a other usbb2k_api process Already run */ 
  fic=open(pid_file,O_RDWR|O_CREAT,0640);
  if (fic<0) 
  {
    syslog(LOG_ERR,"Open PID file failed");
    exit(-1);
  }

  if (lockf(fic,F_TLOCK,0)<0)
  {
    syslog(LOG_ERR,"Error Lockf");
    exit(-1);
  }
  sprintf(pid_str,"%d",getpid());
  write(fic,pid_str,strlen(pid_str));
  
  
  /* Signal catch */
  signal(1, signal_handler);
  signal(15,signal_handler);
  
  return 0; 
}


int 
main(int argc,char *argv[])
{
  int sock_serveur,sock_client,len,t,msgl;
  int ring_type;
  char msg[MSG_MAXL],arg_cmd[MSG_MAXL];
  struct sockaddr_un serveur,client;
  struct usb_device* device;
  pthread_t lect_thread;
  int ret;
  client_arg_t client_arg;

  daemon_mode(); 
  // INIT UNIX_SOCKET
  if ((sock_serveur=socket(AF_UNIX,SOCK_STREAM,0))==-1) {
	syslog(LOG_ERR,"socket");
	exit(1);
   }
  serveur.sun_family=AF_UNIX;
  strcpy(serveur.sun_path,USBB2K_SOCKET);
  unlink(serveur.sun_path);
  len = strlen(serveur.sun_path)+sizeof(serveur.sun_family);
  if (bind(sock_serveur,(struct sockaddr *)&serveur,len) == -1){
    syslog(LOG_ERR,"bind");
    exit(1);
  }

  if (listen(sock_serveur,1)==-1) {
    syslog(LOG_ERR,"listen");
    exit(1);
  }
  
  t = sizeof(client);
  main_loop=1;
  
  do{
    if ((sock_client=accept(sock_serveur,(struct sockaddr*)&client,&t))==-1){
      perror("accept");
      exit(1);
    }
  
    msgl=0;
    client_arg.sock=sock_client;

    // detection du device USB
    device=scan_bus();
    if (device==NULL) {
      sprintf(msg,"USB NOT FOUND\n");
      msgl=write(sock_client,msg,strlen(msg));
      fprintf(stderr,"USB NOT FOUND\n");
      exit(1);
    }
    // Info USB + init
    sprintf(msg,"USB DEVICE FOUND %s\n",device->filename);
    write(sock_client,msg,strlen(msg));
    syslog(LOG_INFO,"Init usb");
    if ((client_arg.dev_h=usbb2k_init(device))==NULL) {
      sprintf(msg,"USB DEVICE INIT ERROR \n");
      fprintf(stderr,msg);
      write(sock_client,msg,strlen(msg));
      close(sock_client);
      continue;
    }
     
    // Thread ECOUTE USB
    if ((ret=pthread_create(&lect_thread,NULL,usblect_th,(void *)&client_arg))!=0){
      fprintf(stderr,"%s\n",strerror(ret));
      exit(1);
    }

    // MAIN LOOP CLIENT (BLOCK PROCESS)
    if (client_th((void *)&client_arg)!=0){
     // Si client disconnect ou USB disconnect 
     usb_loop=OFF;
    }
    usb_loop=OFF; 
    syslog(LOG_INFO,"Client disconnect\n");
    pthread_cancel(lect_thread);
    usleep(50);
    usb_release_interface(client_arg.dev_h,CONTROL);
    //usb_reset(client_arg.dev_h);
    usb_close(client_arg.dev_h);

  }while(main_loop);
  
}
